/*
 * fnt2psf - Converts raw font to PSF
 * Copyright (c) 1999 Martin Hinner <mhi@penguin.cz>
 *
 * Uses part of psfaddtable.c: Add a Unicode character table to a PSF font
 *              Copyright (C) 1994 H. Peter Anvin
 *              Added input ranges, aeb.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "psf.h"

char *help = 
"Usage: %s [-h] [-u unimap] rawfont psfoutfile\n"
"\n"
"Options:\n"
"    -h           Display this help screen\n"
"    -u unimap    Include specified unicode table in PSF\n"
"\n"
"Please report bugs to mhi@penguin.cz\n";

typedef unsigned short unicode;

struct unicode_list
{
  unicode uc;			/* Unicode listed */
  struct unicode_list *next;
};

struct unicode_list **hookptr[512];
struct unicode_list *unilist[512];

void loadunimap(char *filename, int fontlen);


#define EX_DATAERR  2

void usage(char *argv0)
{
    fprintf(stderr,help,argv0);
    exit(1);
}

int getunicode(char **p0)
{
  char *p = *p0;

  while (*p == ' ' || *p == '\t')
    p++;
  if (*p != 'U' || p[1] != '+' ||
      !isxdigit(p[2]) || !isxdigit(p[3]) || !isxdigit(p[4]) ||
      !isxdigit(p[5]) || isxdigit(p[6]))
    return -1;
  *p0 = p+6;
  return strtol(p+2,0,16);
}

void addpair(int fp, int un)
{
  struct unicode_list *newmbr;

  if ( un != PSF_SEPARATOR && un <= 0xFFFF )
    {
	/* Add to linked list IN ORDER READ */
		      
	newmbr = malloc(sizeof(struct unicode_list));
	newmbr->uc = (unicode) un;
	newmbr->next = NULL;
	*(hookptr[fp]) = newmbr;
	hookptr[fp] = &(newmbr->next);
    }
  /* otherwise: ignore */
}

int main(int argc,char **argv)
{
    int ch;
    char *argv0, *unimap;
    int fnt,psf;		/* File handles */
    struct psf_header psfhdr;
    struct stat statb;
    int height;
    void *image;
    struct unicode_list *newmbr, tmpmbr;
    unicode unichar;
    int i;

    argv0 = argv[0];
    unimap = 0;
    height = 0;
    
    while ((ch = getopt (argc, argv, "hu:")) != -1) {
	switch (ch) {
	    case 'h':
		usage(argv0);
	    case 'u':
	    	unimap = (char *)optarg;
		break;
	    default:
		usage(argv0);
	} /* switch */
    } /* while */

    argc -= optind;
    argv += optind;

    if (argc != 2)
	usage(argv0);
      
   if ( (fnt = open(argv[0], O_RDONLY)) < 0 ) {
   	perror(argv[0]);
	exit(1);
   }

   if (fstat(fnt, &statb) < 0) {
   	perror(argv[0]);
	exit(1);
   }

   if ((statb.st_size % 256)!=0) {
   	fprintf(stderr,"%s: Invalid font file\n",argv[0]);
	close(fnt);
	exit(1);
   }

   height = statb.st_size / 256;

   printf("%s: 8x%u font\n",argv[0],height);

   if ((image = malloc(height*256)) == NULL) {
   	fprintf(stderr,"%s: Not enough memory\n",argv[0]);
	close(fnt);
	exit(1);
   }

   if (read(fnt,image,height*256) != height*256) {
   	perror(argv[0]);
	exit(1);
   }
   close(fnt);
 
   psfhdr.magic1 = PSF_MAGIC1;
   psfhdr.magic2 = PSF_MAGIC2;
   psfhdr.mode = 0;
   psfhdr.charsize = height;
   if (unimap != NULL) {
   	psfhdr.mode |= PSF_MODEHASTAB;
	loadunimap(unimap,256);
   }

   if ((psf = creat(argv[1], 0644)) < 0) {
   	perror(argv[1]);
	free(image);
	close(fnt);
	exit(1);
   }

   if (write(psf,&psfhdr,sizeof(psfhdr)) != sizeof(psfhdr) ) {
   	perror(argv[1]);
	exit(1);
   }
 
   if (write(psf,image,height*256) != height*256 ) {
   	perror(argv[1]);
	exit(1);
   }

  unichar = PSF_SEPARATOR;	/* Separator between unicode strings */

  for ( i = 0 ; i < 256 ; i++ )
    {
      for ( newmbr = unilist[i] ; newmbr ; newmbr = tmpmbr.next )
	{
	  tmpmbr = *newmbr;
	  write(psf, &(tmpmbr.uc), sizeof(unicode));
	  free(newmbr);
	}

      /* Write string terminator */
      write(psf, &unichar, sizeof(unicode));
    }

   close(psf);

   return 0;
}


void loadunimap(char *filename, int fontlen)
{
  FILE *ctbl;
  char buffer[65536];
  int i;
  int fp0, fp1, un0, un1;
  char *p, *p1;

  ctbl = fopen(filename, "r");
  if ( !ctbl )
   {
     perror(filename);
     exit(1);
   }

  /* Now we come to the tricky part.  Parse the input table. */

  for ( i = 0 ; i < 512 ; i++ )	/* Initialize unicode list */
    unilist[i] = NULL;

  for ( i = 0 ; i < 512 ; i++ )	/* Initialize hook pointer list */
    hookptr[i] = &(unilist[i]);

  while ( fgets(buffer, sizeof(buffer), ctbl) != NULL )
    {
      if ( (p = strchr(buffer, '\n')) != NULL )
	*p = '\0';
      else
	fprintf(stderr, "%s: Warning: line too long\n", filename);

      p = buffer;

/*
 * Syntax accepted:
 *	<fontpos>	<unicode> <unicode> ...
 *	<range>		idem
 *	<range>		<unicode range>
 *
 * where <range> ::= <fontpos>-<fontpos>
 * and <unicode> ::= U+<h><h><h><h>
 * and <h> ::= <hexadecimal digit>
 */

      while (*p == ' ' || *p == '\t')
	p++;
      if (!*p || *p == '#')
	continue;	/* skip comment or blank line */

      fp0 = strtol(p, &p1, 0);
      if (p1 == p)
	{
	  fprintf(stderr, "Bad input line: %s\n", buffer);
	  exit(EX_DATAERR);
        }
      p = p1;

      while (*p == ' ' || *p == '\t')
	p++;
      if (*p == '-')
	{
	  p++;
	  fp1 = strtol(p, &p1, 0);
	  if (p1 == p)
	    {
	      fprintf(stderr, "Bad input line: %s\n", buffer);
	      exit(EX_DATAERR);
	    }
	  p = p1;
        }
      else
	fp1 = 0;

      if ( fp0 < 0 || fp0 >= fontlen )
	{
	    fprintf(stderr,
		    "%s: Glyph number (0x%x) larger than font length\n",
		    filename, fp0);
	    exit(EX_DATAERR);
	}
      if ( fp1 && (fp1 < fp0 || fp1 >= fontlen) )
	{
	    fprintf(stderr,
		    "%s: Bad end of range (0x%x)\n",
		    filename, fp1);
	    exit(EX_DATAERR);
	}

      if (fp1)
	{
	  /* we have a range; expect the word "idem" or a Unicode range of the
	     same length */
	  while (*p == ' ' || *p == '\t')
	    p++;
	  if (!strncmp(p, "idem", 4))
	    {
	      for (i=fp0; i<=fp1; i++)
		addpair(i,i);
	      p += 4;
	    }
	  else
	    {
	      un0 = getunicode(&p);
	      while (*p == ' ' || *p == '\t')
		p++;
	      if (*p != '-')
		{
		  fprintf(stderr,
"%s: Corresponding to a range of font positions, there should be a Unicode range\n",
			  filename);
		  exit(EX_DATAERR);
	        }
	      p++;
	      un1 = getunicode(&p);
	      if (un0 < 0 || un1 < 0)
		{
		  fprintf(stderr,
"%s: Bad Unicode range corresponding to font position range 0x%x-0x%x\n",
			  filename, fp0, fp1);
		  exit(EX_DATAERR);
	        }
	      if (un1 - un0 != fp1 - fp0)
		{
		  fprintf(stderr,
"%s: Unicode range U+%x-U+%x not of the same length as font position range 0x%x-0x%x\n",
			  filename, un0, un1, fp0, fp1);
		  exit(EX_DATAERR);
	        }
	      for(i=fp0; i<=fp1; i++)
		addpair(i,un0-fp0+i);
	    }
        }
      else
	{
	    /* no range; expect a list of unicode values for a single font position */

	    while ( (un0 = getunicode(&p)) >= 0 )
	      addpair(fp0, un0);
	}
      while (*p == ' ' || *p == '\t')
	p++;
      if (*p && *p != '#')
	fprintf(stderr, "%s: trailing junk (%s) ignored\n", filename, p);
    }

  /* Okay, we hit EOF, now glyph table should be read */

  fclose(ctbl);
}
